/*

   HL_DBValid() : Test a file .dbf validity
   
   Syntax :  HL_DBValid(<cDBF2Vald>) -> nResult
   
   Argument : <cDBF2Vald> : Full file (.dbf) name to test validity
   
   Result : <nResult> : A numeric value to indicate file's validity status :
   
              0 : unidentified file type, highly probably none .dbf
              
             >0 : Validity check successfull; this file is a valid .dbf   
               
             <0 : Error situations : 
             
                  -2        : File not found
                  -nCtrCod  : .dbf with memo, memo file (.dbt or .fpt) not found
                  
                  <-255:
                   Low Level File Error ( - ( FERROR() + 255 ) )

            See DBV_Test.ch for details of <nResult>                    
*/

#include "DBV_Test.ch"

FUNC HL_DBFValid(;                      
                 cDBF2Vald )        // Full file (.dbf) name to test validity
                 
   LOCA nRVal := 0
                  
   LOCA nFilHNo :=   0,;            // File DOS handle number.
        nFerror :=   0,;            // File Error   
        lRetVal := .F.,;            // Return Value
        nHdrLen :=  32             // Length of characters to read from file

   LOCA cHeader  := SPAC(nHdrLen),; // File Header -- Control String
        nReaded  :=  0,;            // Length of characters readed from file
        nCtrCod  :=  0,;            // .dbf validity Control Code
        nLUYear  :=  0,;            // Last Update Year
        nLUMont  :=  0,;            //  "     "    Month
        nLUDay   :=  0,;            //  "     "    Day
        nRECCou  :=  0,;            // Record Count
        n1stDRP  :=  0,;            // Position of first data record
        nFldSRAL :=  0,;            // Field subrecords area length
        nHedLen  :=  0,;            // Header Length (with Fields)
        nRecLen  :=  0,;            // Record Length
        nFldCou  :=  0,;            // Field Count
        nFilSzR  :=  0,;            // Recorded File Size
        nFilSzC  :=  0              // Computed   "    "
        
   LOCA lMemo   :=  .F.,;          // There is M type field ? 
        cStruct := '',;            // File Structure String
        aStruct := {}             // File Structure Array
        
   LOCA cMemoFNm1 := '',;          // Added at 7121 
        cMemoFNm2 := '' 
 
   IF FILE( cDBF2Vald )            // File exists
      IF FILESIZE( cDBF2Vald ) > 0x3F     // 0x20 table hdr + 0x20 1 field record = 0x40
         nFilHNo := FOPEN( cDBF2Vald )     // Open File for Read Only.
         nFerror := FERROR()   
         IF nFerror == 0                 // File open ok.
            nReaded := FREAD( nFilHNo, @cHeader, nHdrLen)
            IF nHdrLen == nReaded               // READ ok.
            
               nCtrCod  :=   ASC( SUBS( cHeader,  1, 1))    // dBF Validity
               nLUYear  :=   ASC( SUBS( cHeader,  2, 1))    // Year of Last Update
               nLUMont  :=   ASC( SUBS( cHeader,  3, 1))    // Month of Last Update
               nLUDay   :=   ASC( SUBS( cHeader,  4, 1))    // Day of Last Update
               nRECCou  := BIN2L( SUBS( cHeader,  5, 4))    // Number of Records
               n1stDRP  := BIN2I( SUBS( cHeader,  9, 2))    // Position of first data record
               nFldSRAL := n1stDRP - ( 32 + IF( nRECCou > 0, 2, 1 ) )  // Field subrecords area length  ( 2/3 ? )
               nRecLen  := BIN2I( SUBS( cHeader, 11, 2))    // Record Length
               nFldCou  := nFldSRAL / 32                   // Field Count
               nFilSzR  := FSEEK( nFilHNo,0,2)             // Recorded File Size
               nFilSzC  := 32 + nFldSRAL + nRecLen * nRecCou      // Computed File Size
               
               IF nLUMont < 13 .AND.;
                  nLUDay  < 32 .AND.;
                  ABS( nFilSzC-nFilSzR ) <= 3         // 3 ?
                  
                  cStruct := SPAC( nFldSRAL )         // File Structure String
                  FSEEK( nFilHNo, 32, 0)            // Go to beg of struct ( field SR area
                  FREAD( nFilHNo, @cStruct, nFldSRAL )   // Read File Structure String
                  
                  HL_DBStruC2A( cStruct, aStruct )

                  lMemo   :=  ASCAN( aStruct, { | a1 | a1[ 2 ] == "M" } ) > 0
                  
                  IF lMemo
                     cMemoFNm1 := ChangFExt( cDBF2Vald, "DBT" )
                     cMemoFNm2 := ChangFExt( cDBF2Vald, "FPT" )
                     IF FILE( cMemoFNm1 ) .OR. FILE( cMemoFNm2 )
                        nRVal := nCtrCod 
                     ELSE
                        nRVal := -nCtrCod  // memo file not found                 
                     ENDIF
                  ELSE   
                     nRVal := nCtrCod 
                  ENDIF
               ELSE  
                  nRVal := 0
               ENDIF nLUMont < 13 ...                   
            ELSE
               nRVal := -30 // nHdrLen # nReaded : Error while reading (Read Fault) 
            ENDIF Read Error
            FCLOSE( nFilHNo )
         ELSE
            nRVal := -( 255 + nFerror ) // LL File Error 
         ENDIF OPEN Error
      ELSE   
         nRVal := -1 // File too small in size
      ENDIF FILESIZE( cDBF2Vald ) > 0x3F   // 0x20 table hdr + 0x20 1 field record = 0x40
   ELSE
      nRVal := -2 // File not found
   ENDIF File existence

RETU nRVal // HL_DBFValid()

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._

PROC HL_DBStruC2A(;                       // dBase Field Structure From String to Array
                cStruct,;  // db file struct in string ( readed low level )
                aStruct )  // db file struct array
                
   LOCA nFldCou := LEN( cStruct ) / 32,;
        c1Field := '',;
        cFldNam := '',;
        cFldTyp := '',;
        nFldWdt :=  0,;
        nFldDec :=  0,;
        nFldNum :=  0
        
   FOR nFldNum := 1 TO nFldCou
      c1Field := SUBS( cStruct, ( nFldNum - 1 ) * 32 + 1, 32 )
      cFldNam := SUBS( c1Field,  1, AT( CHR(0), c1Field ) - 1 )       
      cFldTyp := SUBS( c1Field, 12,  1 ) 
      nFldWdt := ASC( SUBS( c1Field, 17,  1 ) )
      nFldDec := ASC( SUBS( c1Field, 18,  1 ) )
      AADD( aStruct, { cFldNam, cFldTyp, nFldWdt, nFldDec } )
   NEXT nFldNum
   
RETU // HL_DBStruC2A()

*.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._

